﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Text;
using System.ComponentModel;
using System.Windows.Input;
using CrystalDecisions.CrystalReports.Engine;
using System.Collections.ObjectModel;
using System.Windows;
using System.Windows.Forms;
using System.IO;
using System.Windows.Threading;

namespace CrystalReportLocationSetter
{
    public class CrlsViewModel : INotifyPropertyChanged
    {
        #region Class Members

        public event PropertyChangedEventHandler PropertyChanged;
        private string _status, _errorMessage;
        private bool _isVerify;
        private int _currentProgress = 0, _maximumProgress = 100;
        private ReportCollection _reports;
        private CrlsReportDocument _selectedReport;

        #endregion

        #region Properties

        /// <summary>
        /// Construct a new instance of <see cref="CrlsViewModel"/>
        /// </summary>
        public CrlsViewModel()
        {
            _reports = new ReportCollection();
        }

        /// <summary>
        /// The server/DSN name.
        /// </summary>
        public string Server 
        { 
            get { return Settings.Default.Server; }
            set
            {
                Settings.Default.Server = value;
                OnPropertyChanged("Server");
            }
        }

        /// <summary>
        /// The database to which the reports will connect.
        /// </summary>
        public string Database
        {
            get { return Settings.Default.Database; }
            set
            {
                Settings.Default.Database = value;
                OnPropertyChanged("Database");
            }
        }

        /// <summary>
        /// The username for the new connection (required if not using Integrated Security).
        /// </summary>
        public string Username
        {
            get { return Settings.Default.Username; }
            set
            {
                Settings.Default.Username = value;
                OnPropertyChanged("Username");
            }
        }

        /// <summary>
        /// The password for the new connection (required if not using Integrated Security).
        /// </summary>
        public string Password
        {
            get { return Settings.Default.Password; }
            set
            {
                Settings.Default.Password = value;
                OnPropertyChanged("Password");
            }
        }

        /// <summary>
        /// Value indicating whether or not Integrated Security will be used.
        /// </summary>
        public bool IntegratedSecurity
        {
            get { return Settings.Default.IntegratedSecurity; }
            set
            {
                Settings.Default.IntegratedSecurity = value;
                OnPropertyChanged("IntegratedSecurity");
            }
        }

        /// <summary>
        /// If true, the "Verify Database" operation will be performed when "Run" is clicked.
        /// </summary>
        public bool IsVerify
        {
            get { return _isVerify; }
            set
            {
                _isVerify = value;
                OnPropertyChanged("IsVerify");
            }
        }

        /// <summary>
        /// Collection of reports to process.
        /// </summary>
        public ReportCollection Reports
        {
            get { return _reports; }
        }

        /// <summary>
        /// Current status text.
        /// </summary>
        public string Status 
        { 
            get { return _status; }

            set
            {
                _status = value;
                OnPropertyChanged("Status");
            }
        }

        /// <summary>
        /// Progress of the current operation.
        /// </summary>
        public int CurrentProgress
        {
            get { return _currentProgress; }
            set
            {
                _currentProgress = value;
                OnPropertyChanged("CurrentProgress");
            }
        }

        /// <summary>
        /// The currently selected report.
        /// </summary>
        public CrlsReportDocument SelectedReport
        {
            get { return _selectedReport; }
            set
            {
                _selectedReport = value;
                OnPropertyChanged("SelectedReport");
            }
        }

        /// <summary>
        /// Maximum progress that can be reached during the current operation.
        /// </summary>
        public int MaximumProgress
        {
            get { return _maximumProgress; }
            set
            {
                _maximumProgress = value;
                OnPropertyChanged("MaximumProgress");
            }
        }

        protected void OnPropertyChanged(string propertyName)
        {
            var handler = PropertyChanged;
            if (null != PropertyChanged)
            {
                handler(this, new PropertyChangedEventArgs(propertyName));
            }
        }

        #endregion

        #region Commands

        private RelayCommand _setLocationCommand;
        public ICommand SetLocationCommand
        {
            get
            {
                if (_setLocationCommand == null)
                {
                    _setLocationCommand = new RelayCommand(c => this.SetLocation(), c => this.CanSetLocation);
                }

                return _setLocationCommand;
            }
        }

        private RelayCommand _addReportsCommand;
        public ICommand AddReportsCommand
        {
            get
            {
                if (_addReportsCommand == null)
                {
                    _addReportsCommand = new RelayCommand(c => this.AddReports(), c => this.CanAddReports);
                }

                return _addReportsCommand;
            }
        }

        private RelayCommand _removeReportCommand;
        public ICommand RemoveReportCommand
        {
            get
            {
                if (_removeReportCommand == null)
                {
                    _removeReportCommand = new RelayCommand(c => this.RemoveReport(), c => this.CanRemoveReport);
                }

                return _removeReportCommand;
            }
        }

        private RelayCommand _removeAllCommand;
        public ICommand RemoveAllCommand
        {
            get
            {
                if (_removeAllCommand == null)
                {
                    _removeAllCommand = new RelayCommand(c => this.RemoveAll(), c => this.CanRemoveAll);
                }

                return _removeAllCommand;
            }
        }

        private RelayCommand _openReportCommand;
        public ICommand OpenReportCommand
        {
            get
            {
                if (_openReportCommand == null)
                {
                    _openReportCommand = new RelayCommand(c => this.OpenReport(), c => this.CanOpenReport);
                }
                return _openReportCommand;
            }
        }

        private RelayCommand _saveCommand;
        public ICommand SaveCommand
        {
            get
            {
                if (_saveCommand == null)
                {
                    _saveCommand = new RelayCommand(c => this.Save(), c => this.CanSave);
                }
                return _saveCommand;
            }
        }

        /* //TODO: Binding to the menu's exit button isn't working...
        private RelayCommand _exitCommand;
        public ICommand ExitCommand
        {
            get
            {
                if (_exitCommand == null)
                {
                    _exitCommand = new RelayCommand(c =>
                    { 
                        System.Windows.Application.Current.Shutdown(); });
                }
                return _exitCommand;
            }
        }
        */
        #endregion

        #region Public Methods

        /// <summary>
        /// Save the <see cref="ReportCollection"/> to a specified location.
        /// </summary>
        public void Save()
        {
            var dialog = new System.Windows.Forms.FolderBrowserDialog();
            dialog.ShowNewFolderButton = true;
            if (DialogResult.OK == dialog.ShowDialog())
            {
                if (!String.IsNullOrEmpty(dialog.SelectedPath))
                    saveAsync(dialog.SelectedPath);
            }
        }

        /// <summary>
        /// Opens <see cref="SelectedReport"/> in the registered program.
        /// </summary>
        public void OpenReport()
        {
            Process.Start(this.SelectedReport.FileName);
        }

        /// <summary>
        /// Removes a report from <see cref="ReportCollection"/>
        /// </summary>
        public void RemoveReport()
        {
            if (MessageBoxResult.Yes == System.Windows.MessageBox.Show("Do you realy want to remove the selected report?", "Remove Confirmation", System.Windows.MessageBoxButton.YesNo))
            {
                var doc = SelectedReport;
                Reports.Remove(doc);
                doc.Close();
                doc.Dispose();
                doc = null;
            }
        }

        /// <summary>
        /// Removes all reports from <see cref="ReportCollection"/>
        /// </summary>
        public void RemoveAll()
        {
            if (MessageBoxResult.Yes == System.Windows.MessageBox.Show("Do you realy want to remove all reports?", "Remove all reports?", System.Windows.MessageBoxButton.YesNo))
            {
                MaximumProgress = Reports.Count();

                var worker = new BackgroundWorker();
                worker.DoWork += (s, e) =>
                {
                    for(int i = 0; i < Reports.Count(); i++)
                    {
                        var report = Reports[i];
                        
                        try
                        {
                            Status = String.Format("Removing {0}...", report.FileName);
                            report.Close();
                            report.Dispose();
                        }
                        catch (Exception ex)
                        {
                            Status = ex.ToString();
                        }
                    }

                    Reports.Clear();
                    Status = "Finished removing reports.";
                    CurrentProgress = 0;
                };
            }
        }

        /// <summary>
        /// Adds selected reports to <see cref="ReportCollection"/>
        /// </summary>
        public void AddReports()
        {
            var ofd = new Microsoft.Win32.OpenFileDialog();
            ofd.Multiselect = true;
            
            if (ofd.ShowDialog() == true)
            {
                string[] filePath = ofd.FileNames;
                string[] safeFilePath = ofd.SafeFileNames;
                addReportsAsync(filePath);
            }
        }

        /// <summary>
        /// Performs the "Set datasource location" process on each report in <see cref="ReportCollection"/>
        /// using the entered connection information.
        /// </summary>
        public void SetLocation()
        {
            if (!IsValid)
            {
                System.Windows.MessageBox.Show(_errorMessage);
                return;
            }

            var dto = new ReportConnectionDto(Server, Database, Username, Password, IntegratedSecurity);
            ReportService reportService = new ReportService(dto);
            MaximumProgress = Reports.Count;
            reportService.SetLocationProgressChanged += (s, e) => { CurrentProgress = e.Data; };
            reportService.SetLocationComplete += (s, e) => 
            { 
                CurrentProgress = 0; 
                MaximumProgress = 100;
                OnPropertyChanged("Reports");
                Reports.NotifyChanged();
            };
            reportService.SetLocationForReportsAsync(Reports, false);
        }

        #endregion Public Methods

        #region Private Methods

        private List<CrlsReportDocument> _addList = new List<CrlsReportDocument>();

        private void addReportsAsync(string[] files)
        {
            if (null == files) return;

            _addList = new List<CrlsReportDocument>();

            BackgroundWorker worker = new BackgroundWorker();
            worker.WorkerReportsProgress = true;
            worker.ProgressChanged += (s, e) => { CurrentProgress = e.ProgressPercentage; };
            worker.RunWorkerCompleted += (s, e) => { _addList.ForEach(c => Reports.Add(c)); };
            worker.DoWork += (s, e) =>
            {
                var bgworker = s as BackgroundWorker;
                bgworker.ReportProgress(0, "Starting save process...");
                MaximumProgress = files.Count();

                for (int i = 0; i < files.Count(); i++)
                {
                    string file = files[i];
                    bgworker.ReportProgress(i);

                    try
                    {
                        Status = String.Format("Loading {0}...", file);
                        var doc = new CrlsReportDocument();
                        doc.Load(file);
                        _addList.Add(doc);
                    }
                    catch(Exception error)
                    {
                        Status = error.ToString();
                    }
                }

                Status = "Finished adding reports!";
                bgworker.ReportProgress(100);
            };
            worker.RunWorkerAsync();
        }

        private void saveAsync(string path)
        {
            if (null == Reports || Reports.Count == 0) return;

            BackgroundWorker worker = new BackgroundWorker();
            worker.WorkerReportsProgress = true;
            worker.RunWorkerCompleted += (s, e) => { };
            worker.ProgressChanged += (s, e) => { CurrentProgress = e.ProgressPercentage; };
            worker.DoWork += (s, e) =>
            {
                MaximumProgress = Reports.Count;
                var bgworker = s as BackgroundWorker;
                bgworker.ReportProgress(0, "Starting save process...");

                for(int i = 0; i < Reports.Count; i++)
                {
                    bgworker.ReportProgress(i);
                    var report = Reports[i];

                    try
                    {
                        string newName = Path.Combine(path, Path.GetFileName(report.FileName));
                        Status = String.Format("Saving {0}...", newName);
                        report.SaveAs(newName, false);
                    }
                    catch (Exception ex)
                    {
                        Status = ex.ToString();
                    }
                }

                Status = "Finished saving reports!";
                bgworker.ReportProgress(MaximumProgress);
            };
            worker.RunWorkerAsync();
        }

        #endregion

        #region Public Properties

        public bool CanSave
        {
            get { return (null != Reports) && (Reports.Count > 0); }
        }

        public bool CanSetLocation
        {
            get { return (null != Reports) && (Reports.Count > 0); }
        }

        public bool CanAddReports
        {
            get { return true; }
        }

        public bool CanRemoveAll
        {
            get { return (null != Reports && Reports.Count > 0); }
        }

        public bool CanRemoveReport
        {
            get { return (null != SelectedReport); }
        }

        public bool CanOpenReport { get { return null != SelectedReport; } }

        public bool IsValid
        {
            get
            {
                OnPropertyChanged("IsValid");
                return Validate();
            }
        }

        public bool Validate()
        {
            _errorMessage = String.Empty;
            bool result = true;

            if (!IntegratedSecurity && String.IsNullOrEmpty(Username))
            {
                _errorMessage += "* Username must have a value when not using Integrated Security." + Environment.NewLine;
                result = false;
            }
            
            if (!IntegratedSecurity && String.IsNullOrEmpty(Password))
            {
                _errorMessage += "* Password must have a value when not using Integrated Security." + Environment.NewLine;
                result = false;
            }

            if (String.IsNullOrEmpty(Server))
            {
                _errorMessage += "* Server must have a value." + Environment.NewLine;
                result = false;
            }

            if (String.IsNullOrEmpty(Database))
            {
                _errorMessage += "* Database must have a value." + Environment.NewLine;
                result = false;
            }

            return result;
        }

        #endregion
    }
}
