﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using CrystalDecisions.CrystalReports.Engine;
using System.IO;
using CrystalDecisions.Shared;
using System.ComponentModel;
using NLog;
using System.Windows.Threading;
using System.Windows;


namespace CrystalReportLocationSetter
{
    /// <summary>
    /// <see cref="ReportService"/> encapsulates the methods required to manage connection
    /// information and other report specific tasks.
    /// </summary>
    public class ReportService
    {
        private static Logger _logger = LogManager.GetCurrentClassLogger();

        /// <summary>
        /// Gets or sets the <see cref="ReportConnectionDto"/> used by the service.
        /// </summary>
        public ReportConnectionDto ReportConnectionDto { get; set; }

        /// <summary>
        /// Event fires when the service's progress changes during an async operation.
        /// </summary>
        public event EventHandler<EventArgs<int>> SetLocationProgressChanged;

        /// <summary>
        /// Event fires when the service completes an async operation.
        /// </summary>
        public event EventHandler SetLocationComplete;

        /// <summary>
        /// Creates a new instance of <see cref="ReportService"/>
        /// </summary>
        /// <param name="reportConnectionDto">The <see cref="ReportConnectionDto"/> the service will use to
        /// update report connection information.
        /// </param>
        public ReportService(ReportConnectionDto reportConnectionDto)
        {
           ReportConnectionDto = reportConnectionDto;
        }

        /// <summary>
        /// Asynchronously performs the "Set Datasource Location" process on the collection of reports.
        /// </summary>
        /// <param name="reports">The collection of reports to process.</param>
        /// <param name="isVerify">If true, "Verify Database" will also be performed on the reports.</param>
        public void SetLocationForReportsAsync(ReportCollection reports, bool isVerify)
        {
            BackgroundWorker worker = new BackgroundWorker();
            worker.WorkerReportsProgress = true;
            worker.RunWorkerCompleted += (s, e) => { onSetLocationComplete(); };
            worker.ProgressChanged += (s, e) => { onSetLocationProgressChanged(e.ProgressPercentage);  };
            worker.DoWork += (s, e) => 
            {
                var bgworker = s as BackgroundWorker;
                bgworker.ReportProgress(0, "Starting set location process...");

                for (int i = 0; i < reports.Count; i++ )
                {
                    var report = reports[i];

                    try
                    {
                        SetLocationForReport(report, isVerify);
                    }
                    catch (Exception ex)
                    {
                        _logger.LogException(LogLevel.Error, "Set location failed.", ex);
                    }
                    bgworker.ReportProgress(i);
                }
            };

            worker.RunWorkerAsync();
        }

        /// <summary>
        /// Performs the "Set Datasource Location" process on a single report.
        /// </summary>
        /// <param name="reports">The report to process.</param>
        /// <param name="isVerify">If true, "Verify Database" will also be performed on the report.</param>
        public void SetLocationForReport(CrlsReportDocument doc, bool isVerify)
        {
            udpateConnectionInfo(doc);
  
            if (isVerify)
            {
              doc.VerifyDatabase();
            }
        }

        #region Private Methods

        private void onSetLocationProgressChanged(int progress)
        {
            var handler = SetLocationProgressChanged;
            if (handler != null)
            {
                handler(this, new EventArgs<int>() { Data = progress });
            }
        }

        private void onSetLocationComplete()
        {
            var handler = SetLocationComplete;
            if (handler != null)
            {
                handler(this, new EventArgs());
            }
        }

        private void udpateConnectionInfo(CrlsReportDocument reportDocument)
        {
            ConnectionInfo connectionInfo = new ConnectionInfo();
            connectionInfo.ServerName = ReportConnectionDto.Server;
            connectionInfo.DatabaseName = ReportConnectionDto.Database;
            connectionInfo.UserID = ReportConnectionDto.Username;
            connectionInfo.Password = ReportConnectionDto.Password;

            setDataSourceLocation(connectionInfo, reportDocument);
            setDataSourceLocationForSubreports(connectionInfo, reportDocument);
        }

        private void setDataSourceLocation(ConnectionInfo connectionInfo, ReportDocument pCrystalReportDocument)
        {
            foreach (IConnectionInfo connectInfo in pCrystalReportDocument.DataSourceConnections)
            {
                connectInfo.SetConnection(connectionInfo.ServerName, connectionInfo.DatabaseName,
                    connectionInfo.UserID, connectionInfo.Password);
            }

            foreach (Table table in pCrystalReportDocument.Database.Tables)
            {
                TableLogOnInfo tableLogonInfo = table.LogOnInfo;
                tableLogonInfo.ConnectionInfo = connectionInfo;
                table.ApplyLogOnInfo(tableLogonInfo);
                table.Location = table.LogOnInfo.TableName;
            }
        }

        private void setDataSourceLocationForSubreports(ConnectionInfo connectionInfo, CrlsReportDocument reportDocument)
        {
            if (reportDocument.Subreports == null) return;

            foreach (ReportDocument subreport in reportDocument.Subreports)
            {
                setDataSourceLocation(connectionInfo, subreport);
            }
        }

        #endregion

    }
}
